;
;
;     YLISP
;
;   () 2010  ..
;


(defconstant *delay* 0.03) ;   -
(defvar *remstep* 10) ;     -  

(defvar *bgcolor* #x70)

(defvar *win* nil) ; ,    

(setf *print-pretty* nil)

(defvar *score* 0)

;  -   
(defstruct pos y x) ;   ,
                    ;     

;  -> 
(defun convert-pos (l)
  (make-pos :y (first l) :x (second l)))

;  ->  
(defun make-figtype (l)
    (mapcar #'convert-pos l))

;  -    4    
(defconstant *figure-types* 
  (make-array 3 :initial-contents 
    (mapcar #'make-figtype '(((0 0) (-1 0) (1 0) (-2 0))
                             ((0 0) (-1 0) (0 -1) (1 0))
                             ((0 0) (-1 0) (1 0) (1 1))))))

;  
(defmacro p (y x)
  `(make-pos :y ,y :x ,x))

;,    
(defconstant *start-pos* (p 2 5))

(defun rot-clock (pos) ;   0, 0  
  (make-pos :y (pos-x pos) :x (- (pos-y pos))))

(defun rot-aclock (pos) ;   0, 0  
  (make-pos :y (- (pos-x pos)) :x (pos-y pos)))

(defstruct figure type color)
(defvar *figure*)

(defvar *cupsize* (p 24 11)) ; 4   

;
(defvar *cup* (make-array (list (pos-y *cupsize*) 
                                (pos-x *cupsize*)) 
                          :initial-element *bgcolor*))

(defvar *strings* (make-hash-table))
; ,    
;,   ,  
(defun get-sop (length)
  (multiple-value-bind (val found) (gethash length *strings*)
    (if found val
      (setf (gethash length *strings*) 
            (make-string length :initial-element #\Space)))))

(defun draw-cup ()
  (dotimes (i (pos-y *cupsize*))
    (when (> i 3) ;  4   ,   
      (do  ((end 1) (start 0) 
            (starcol (aref *cup* i 0)) (ecol))
           ((= start (pos-x *cupsize*)))
        (if (and (< end (pos-x *cupsize*))
                 (= starcol (setf ecol (aref *cup* i end))))
          (incf end)
          (progn
            (setf (window-color *win*) starcol
                  (cursor-position *win*) (vector (- i 4) (* 2 start)))
            (princ #++ (make-string (- end start) :initial-element #\Space) 
                       (get-sop (* 2 (- end start))) *win*)
            (setf start end end (1+ end) starcol ecol)))))))

(defun pos-add (p1 p2)
  (make-pos :y (+ (pos-y p1) (pos-y p2)) 
            :x (+ (pos-x p1) (pos-x p2))))

(defmacro cup-elt (pos)
  `(aref *cup* (pos-y ,pos) (pos-x ,pos)))

(defun put-point (point color)
  (setf (aref *cup* (pos-y point) (pos-x point)) color))

(defun draw-point (point color)
  (when (>= (pos-y point) 4)
    (setf (cursor-position *win*) 
          (vector (- (pos-y point) 4) (* 2 (pos-x point)))
          (window-color *win*) color)
    (princ "  " *win*)))

(defun diff-draw-figure (posl1 posl2 drawfunc color)
;         posl2, 
;   posl1 (  )
;
;"posl" - position list
  (dolist (p posl1)
    (unless (member p posl2 :test #'equal)
      (funcall drawfunc p *bgcolor*)))
  (dolist (p posl2)
    (unless (member p posl1 :test #'equal)
      (funcall drawfunc p color))))

; -  -
(defun draw-figure (figure position drawfunc erasep)
  (dolist (elt (figure-type figure))
    (funcall drawfunc (pos-add position elt) 
      (if erasep *bgcolor* (figure-color figure)))))

;?
(defun fig-collidep (figure position)
   (do* ;  ,   
       ((typelist (cons nil (figure-type figure)) (cdr typelist))
        (coll-this nil (and typelist 
                            (not (= (cup-elt (pos-add (car typelist) 
                                                      position)) 
                                    *bgcolor*)))))
     ((or (null typelist) coll-this) coll-this)))

; ?
(defun fig-outbound (figure position)
  (do* ((typelist (cons nil (figure-type figure)) (cdr typelist))
        (new-pos nil (and typelist 
                          (pos-add position (car typelist))))
        (outbound nil (and new-pos
                          (or (< (pos-x new-pos) 0)
                          (>= (pos-x new-pos) (pos-x *cupsize*))
                          (< (pos-y new-pos) 0)
                          (>= (pos-y new-pos) (pos-y *cupsize*))))))
    ((or (null typelist) outbound) outbound)))

; ,   ?
(defun fig-fit (figure position)
  (not (or (fig-outbound figure position)
           (fig-collidep figure position))))

(defun drop-all () ;     
  (dotimes (ii (pos-y *cupsize*))
    (let ((i (- (pos-y *cupsize*) ii 1)))
      (if (zerop i) 
        (dotimes (j (pos-x *cupsize*))
          (setf (aref *cup* i j) *bgcolor*))
        (dotimes (j (pos-x *cupsize*))
            (setf (aref *cup* i j) (aref *cup* (1- i) j)))))))


(defun drop-straight (i j) ;  
  (setf (aref *cup* i j)
    (if (zerop i) *bgcolor*
      (let ((ucell (aref *cup* (1- i) j))) 
        (drop-straight (1- i) j)
        ucell))))

(defun drop-col (i j) ;   ,   
  (unless (zerop i)   ;     (0..i j),  NIL
    (if (= (aref *cup* i j) (aref *cup* (1- i) j))
      (drop-col (1- i) j)
      (if (= (aref *cup* i j) *bgcolor*)
        (drop-straight i j)  ;  NIL
        (drop-col (1- i) j)))))


(defun drop-bricks () ;    
  (do ()
    ((zerop 
       (loop for j from 0 to (1- (pos-x *cupsize*))
         count (drop-col (1- (pos-y *cupsize*)) j)))) (draw-cup)))

(defun cut-line (i)
  (dotimes (j (1- (pos-x *cupsize*)))
    (drop-straight i j)))

;1.   
;2.      
(defun check-n-drop ()
  ;  
  (loop for i from (1- (pos-y *cupsize*)) downto 0
        do (when (loop for j from 0 to (1- (pos-x *cupsize*)) 
                       never (= (cup-elt (p i j)) *bgcolor*))
               (cut-line i) (incf *score*)
               (draw-cup))))

(defun new-figure()
   (setf *figure* (make-figure 
                      :type (aref *figure-types* (random 3)) 
                      :color (do ((color *bgcolor* 
                                      (* 16 (random 16))))
                                 ((/= *bgcolor* color) color)))
         *figure-pos* *start-pos*))

(defun cycle(i) 
  (let ((l (cons i nil)))
    (rplacd l l)))

(defun draw-score ()
  ; 
  (setf (cursor-position *terminal-io*) (vector 5 40)
        (window-color *terminal-io*) #x60)
  (format t ": ~D    " (* *score* 100)))

(defun tetris ()
  (let ((*win* (make-window (vector 2 10 (+ 2 -4 (pos-y *cupsize*)) 
                                     (+ 10 (* 2 (pos-x *cupsize*)))) 
                        *bgcolor*))
        (oldcolor (window-color *terminal-io*)))


    (setf (window-color *terminal-io*) #x16)
    (window-shot *terminal-io*)
    (clear-window *terminal-io*)
    (set-cursor-shape nil)
    (draw-cup) (draw-score)
    (new-figure)
    (format *status-output* "~% YLISP-.   .")
    (read-char)
    (format *status-output* "~% YLISP-. ")
    (do ((key nil (read-char-no-hang))
         (itno 0 (1+ itno))
         (gameover nil))
        ((or (eq key #\ESC) gameover))

      (when (eq key #\UP) ;  
        (let* ((newtype (mapcar #'rot-aclock (figure-type *figure*)))
               ; ,    
               (newfigure (make-figure :type newtype 
                                       :color (figure-color *figure*)))
               ;   mapcar
               (posl (cycle *figure-pos*)))
          (rplacd posl posl)
          (when (fig-fit newfigure *figure-pos*)
            (let ((oldposl (mapcar #'pos-add posl (figure-type *figure*)))
                  (newposl (mapcar #'pos-add posl newtype)))
            (diff-draw-figure oldposl newposl 
                     #'draw-point (figure-color *figure*))
            (setf *figure* newfigure)))))

      (when (eq key #\DOWN)
        (setf *remstep* 1))

      (let* ((movpos1 (if (zerop (rem itno *remstep*)) 
                         (pos-add *figure-pos* (p 1 0)) 
                                  *figure-pos*)) ;
             (movpos2 (pos-add ;       
                           (cond ((eq key #\LEFT) (p 0 -1))
                                 ((eq key #\RIGHT) (p 0 1))
                                 (t (p 0 0)))
                        movpos1))
             (ff1 (fig-fit *figure* movpos1))
             (ff2 (fig-fit *figure* movpos2))
             (sidecol (and ff1 (not ff2)))
             (nogo (and (not ff1) (not ff2)))
             (newpos (cond (nogo *figure-pos*)
                           (sidecol movpos1)
                           (t movpos2))))
         (setf (cursor-position *terminal-io*) (vector 20 0))
 ;        (format t "MP1 ~A FF1 ~5A ~T MP2 ~A FF2 ~5A C ~X BC ~X  " 
 ;          movpos1 ff1 movpos2 ff2 (figure-color *figure*) *bgcolor*)
 ;        (unless (and ff1 ff2) (sleep 0.5))
         (if nogo
           (if (eql *figure-pos* *start-pos*)
             (progn 
               (setf gameover t)
               (format *status-output* "~% ! "))
             (progn ;  -     
               (diff-draw-figure nil (mapcar #'pos-add 
                                           (figure-type *figure*) 
                                           (cycle *figure-pos*))
                            #'put-point (figure-color *figure*))
               ;  
               (check-n-drop)
               ;  
               (new-figure)
               ; 
               (draw-score)
               ;    
               (setf *remstep* 10)))
           ; 
           (progn
             (diff-draw-figure (mapcar #'pos-add 
                                  (figure-type *figure*) 
                                  (cycle *figure-pos*))
                               (mapcar #'pos-add 
                                  (figure-type *figure*) 
                                  (cycle newpos))
                      #'draw-point (figure-color *figure*))
             (setf *figure-pos* newpos))))
         (sleep *delay*))
    (format *status-output* "  .")
    (read-char)
    (window-restore *terminal-io*)
    (setf (window-color *terminal-io*) oldcolor)
    (set-cursor-shape t)
    (close *win*)))

(tetris)


